package org.eclipse.emf.cdo.server.internal.db.mapping.horizontal;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.runtime.Assert;
import org.eclipse.emf.cdo.common.branch.CDOBranch;
import org.eclipse.emf.cdo.common.id.CDOID;
import org.eclipse.emf.cdo.common.revision.CDORevision;
import org.eclipse.emf.cdo.common.revision.delta.CDOListFeatureDelta;
import org.eclipse.emf.cdo.eresource.EresourcePackage;
import org.eclipse.emf.cdo.server.IStoreAccessor.QueryXRefsContext;
import org.eclipse.emf.cdo.server.IStoreChunkReader.Chunk;
import org.eclipse.emf.cdo.server.db.IDBStoreAccessor;
import org.eclipse.emf.cdo.server.db.IDBStoreChunkReader;
import org.eclipse.emf.cdo.server.db.IPreparedStatementCache;
import org.eclipse.emf.cdo.server.db.IPreparedStatementCache.ReuseProbability;
import org.eclipse.emf.cdo.server.db.mapping.IListMappingDeltaSupport;
import org.eclipse.emf.cdo.server.db.mapping.IMappingStrategy;
import org.eclipse.emf.cdo.server.internal.db.CDODBSchema;
import org.eclipse.emf.cdo.server.internal.db.bundle.OM;
import org.eclipse.emf.cdo.spi.common.revision.InternalCDORevision;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.net4j.db.DBException;
import org.eclipse.net4j.db.DBUtil;
import org.eclipse.net4j.db.ddl.IDBTable;
import org.eclipse.net4j.util.collection.MoveableList;
import org.eclipse.net4j.util.om.trace.ContextTracer;

/**
 * @author Christophe
 * 
 */
public class NonAuditListTableMapping2 extends BasicAbstractListTableMapping implements IListMappingDeltaSupport {
    private static final ContextTracer TRACER=new ContextTracer(OM.DEBUG, NonAuditListTableMapping2.class);

    private String sqlSelectChunksPrefix;

    private NonAuditListTableMapping delegate;

    public NonAuditListTableMapping2(IMappingStrategy mappingStrategy, EClass eClass, EStructuralFeature feature) {
        super(mappingStrategy, eClass, feature);
        if (feature instanceof EReference && ((EReference) getFeature()).isContainment()
                && !feature.equals(EresourcePackage.eINSTANCE.getCDOResource_Contents())
                && !feature.equals(EresourcePackage.eINSTANCE.getCDOResourceFolder_Nodes()))
            initSQLStrings();
        else
            delegate=new NonAuditListTableMapping(mappingStrategy, eClass, feature);
    }

    @Override
    public void addSimpleChunkWhere(IDBStoreAccessor accessor, CDOID cdoid, StringBuilder builder, int index) {
        if (delegate == null) {
            builder.append(" LIMIT "); //$NON-NLS-1$
            builder.append(index);
            builder.append(",1"); //$NON-NLS-1$
        } else
            delegate.addSimpleChunkWhere(accessor, cdoid, builder, index);
    }

    @Override
    public void addRangedChunkWhere(IDBStoreAccessor accessor, CDOID cdoid, StringBuilder builder, int fromIndex, int toIndex) {
        if (delegate == null) {
            builder.append(" LIMIT "); //$NON-NLS-1$
            builder.append(fromIndex);
            builder.append(',');
            builder.append(toIndex - fromIndex);
        } else
            delegate.addRangedChunkWhere(accessor, cdoid, builder, fromIndex, toIndex);
    }

    private void initSQLStrings() {
        // List<EClass> candidates=new ArrayList<>();
        // if (getFeature().getEType() instanceof EClass) {
        // EClass type=(EClass) getFeature().getEType();
        // for (EClassifier classifier : ModelPackage.eINSTANCE.getEClassifiers()) {
        // if (classifier instanceof EClass) {
        // EClass clazz=(EClass) classifier;
        // if (type.isSuperTypeOf(clazz) && !clazz.isAbstract())
        // candidates.add(clazz);
        // }
        // }
        // }
        // if (candidates.size() != 1)
        //            TRACER.trace("WARNING: There must be only one candidate for chunking for feature " + getFeature()); //$NON-NLS-1$
        EClass clazz=(EClass) getFeature().getEType();
        Assert.isTrue(!clazz.isAbstract(), "Type of the reference must not be abstract " + getFeature());

        // sqlSelectChunksPrefix=new String[candidates.size()];

        // ---------------- SELECT to read chunks ----------------------------
        int i=0;
        String colName;
        if (getFeature().equals(EresourcePackage.eINSTANCE.getCDOResource_Contents()))
            colName=CDODBSchema.ATTRIBUTES_RESOURCE;
        else
            colName=CDODBSchema.ATTRIBUTES_CONTAINER;

        // for (EClass candidate : candidates) {
            StringBuilder builder=new StringBuilder();
            builder.append("SELECT "); //$NON-NLS-1$
            builder.append(CDODBSchema.ATTRIBUTES_ID);
            builder.append(" FROM "); //$NON-NLS-1$
        builder.append(getMappingStrategy().getTableName(clazz));
            builder.append(" WHERE "); //$NON-NLS-1$

            builder.append(colName);
            builder.append("=? AND cdo_version>0"); //$NON-NLS-1$

        sqlSelectChunksPrefix=builder.toString();
        // }
    }

    @Override
    public Collection<IDBTable> getDBTables() {
        if (delegate == null)
            return Collections.emptyList();
        else
            return delegate.getDBTables();
    }

    @Override
    public void objectDetached(IDBStoreAccessor accessor, CDOID id, long revised) {
        if (delegate == null) {
            // NO-OP
        } else {
            delegate.objectDetached(accessor, id, revised);
        }
    }

    // @Override
    protected void setKeyFields(PreparedStatement stmt, CDORevision revision) throws SQLException {
        getMappingStrategy().getStore().getIDHandler().setCDOID(stmt, 1, revision.getID());
    }

    @Override
    public void writeValues(IDBStoreAccessor accessor, InternalCDORevision revision) {
        if (delegate == null) {
            // NO-OP
        } else {
            delegate.writeValues(accessor, revision);
        }
    }

    @Override
    public void readValues(IDBStoreAccessor accessor, InternalCDORevision revision, int listChunk) {
        if (delegate == null) {
            MoveableList<Object> list=revision.getList(getFeature());

            if (listChunk == 0 || list.size() == 0) {
                // nothing to read take shortcut
                return;
            }

            if (TRACER.isEnabled()) {
                TRACER.format("Reading list values for feature {0}.{1} of {2}v{3}", getContainingClass().getName(), //$NON-NLS-1$
                        getFeature().getName(), revision.getID(), revision.getVersion());
            }

            IPreparedStatementCache statementCache=accessor.getStatementCache();
            PreparedStatement stmt=null;
            ResultSet resultSet=null;

            // for (String sql : sqlSelectChunksPrefix) {
            try {
                // ((EClass)getFeature().getEType())
                // String sql=sqlSelectChunksPrefix + sqlOrderByIndex;
                stmt=statementCache.getPreparedStatement(sqlSelectChunksPrefix, ReuseProbability.HIGH);
                setKeyFields(stmt, revision);

                if (TRACER.isEnabled()) {
                    TRACER.trace(stmt.toString());
                }

                if (listChunk != CDORevision.UNCHUNKED) {
                    stmt.setMaxRows(listChunk); // optimization - don't read unneeded rows.
                }

                resultSet=stmt.executeQuery();

                int currentIndex=0;
                while ((listChunk == CDORevision.UNCHUNKED || --listChunk >= 0) && resultSet.next()) {
                    Object value=getValue(resultSet);
                    if (TRACER.isEnabled()) {
                        TRACER.format("Read value for index {0} from result set: {1}", list.size(), value); //$NON-NLS-1$
                    }

                    list.set(currentIndex++, value);
                }
            } catch (SQLException ex) {
                throw new DBException(ex);
            } finally {
                DBUtil.close(resultSet);
                statementCache.releasePreparedStatement(stmt);
            }
            // }

            if (TRACER.isEnabled()) {
                TRACER.format("Reading list values done for feature {0}.{1} of {2}v{3}", getContainingClass().getName(), //$NON-NLS-1$
                        getFeature().getName(), revision.getID(), revision.getVersion());
            }
        } else {
            delegate.readValues(accessor, revision, listChunk);
        }
    }

    /**
     * @deprecated
     */
    @Deprecated
    protected CDOID getValue(ResultSet resultSet) throws SQLException {
        return getMappingStrategy().getStore().getIDHandler().getCDOID(resultSet, 1);
    }

    @Override
    public void readChunks(IDBStoreChunkReader chunkReader, List<Chunk> chunks, String where) {
        if (delegate == null) {
            if (TRACER.isEnabled()) {
                TRACER.format("Reading list chunk values for feature {0}.{1} of {2}v{3}", getContainingClass().getName(), //$NON-NLS-1$
                        getFeature().getName(), chunkReader.getRevision().getID(), chunkReader.getRevision().getVersion());
            }

            IPreparedStatementCache statementCache=chunkReader.getAccessor().getStatementCache();
            PreparedStatement stmt=null;
            ResultSet resultSet=null;

            Chunk chunk=null;
            int chunkSize=0;
            int chunkIndex=0;
            int indexInChunk=0;

            // for (String sql : sqlSelectChunksPrefix) {
            try {
                StringBuilder builder=new StringBuilder(sqlSelectChunksPrefix);
                builder.append(where);

                stmt=statementCache.getPreparedStatement(builder.toString(), ReuseProbability.LOW);
                setKeyFields(stmt, chunkReader.getRevision());

                resultSet=stmt.executeQuery();

                while (resultSet.next()) {
                    Object value=getValue(resultSet);

                    if (chunk == null) {
                        chunk=chunks.get(chunkIndex++);
                        chunkSize=chunk.size();

                        if (TRACER.isEnabled()) {
                            TRACER.format("Current chunk no. {0} is [start = {1}, size = {2}]", chunkIndex - 1, chunk.getStartIndex(), //$NON-NLS-1$
                                    chunkSize);
                        }
                    }

                    if (TRACER.isEnabled()) {
                        TRACER.format("Read value for chunk index {0} from result set: {1}", indexInChunk, value); //$NON-NLS-1$
                    }

                    chunk.add(indexInChunk++, value);
                    if (indexInChunk == chunkSize) {
                        if (TRACER.isEnabled()) {
                            TRACER.format("Chunk finished"); //$NON-NLS-1$
                        }

                        chunk=null;
                        indexInChunk=0;
                    }
                }

                if (TRACER.isEnabled()) {
                    TRACER.format("Reading list chunk values done for feature {0}.{1} of {2}v{3}", getContainingClass().getName(), //$NON-NLS-1$
                            getFeature().getName(), chunkReader.getRevision().getID(), chunkReader.getRevision().getVersion());
                }
            } catch (SQLException ex) {
                throw new DBException(ex);
            } finally {
                DBUtil.close(resultSet);
                statementCache.releasePreparedStatement(stmt);
            }
            // }
        } else
            delegate.readChunks(chunkReader, chunks, where);
    }

    @Override
    public boolean queryXRefs(IDBStoreAccessor accessor, String mainTableName, String mainTableWhere, QueryXRefsContext context, String idString) {
        if (delegate == null) {
            // TODO Auto-generated method stub
            return false;
        } else {
            return delegate.queryXRefs(accessor, mainTableName, mainTableWhere, context, idString);
        }
    }

    public void rawDeleted(IDBStoreAccessor accessor, CDOID id, CDOBranch branch, int version) {
    }

    @Override
    public void processDelta(IDBStoreAccessor accessor, CDOID id, int branchId, int oldVersion, int newVersion, long created, CDOListFeatureDelta delta) {
        if (delegate == null) {
            // NO-OP
        } else {
            delegate.processDelta(accessor, id, branchId, oldVersion, newVersion, created, delta);
        }
    }
}
